/*
 * Copyright (C) 2008 Intel Corporation.
 * All rights reserved
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, 
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR 
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, 
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, 
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.  
 *
 */


#ifndef IPMI_DCMI_H
#define IPMI_DCMI_H

#include <ipmitool/ipmi.h>

/* DCMI commands per DCMI 1.0 SPEC */

#define DCMI_GROUP         		0xDC  /* Group Extension Identification */
#define DCMI_CMD_CAPABILITY   0x01
#define DCMI_CMD_GETRED    		0x02
#define DCMI_CMD_GETLMT    		0x03
#define DCMI_CMD_SETLMT    		0x04
#define DCMI_CMD_PWRACT    		0x05
#define DCMI_CMD_GETASSET  		0x06
#define DCMI_CMD_GETSNSR   		0x07
#define DCMI_CMD_PWRMGT    		0x08
#define DCMI_CMD_CONFORM   		0x0001

struct dcmi_cmd {
    uint16_t val;
    const char * str;
    const char * desc;
};

/* make a struct for the return from the get limit command */
struct power_limit {
    uint8_t grp_id; /* first byte: Group Extension ID */
    uint16_t reserved_1; /* second and third bytes are reserved */
    uint8_t action; /* fourth byte is the exception action */
    uint16_t limit; /* fifth through sixth byte are the power limit in watts */
    uint32_t correction; /* seventh - 10th bytes are the correction period */
    uint16_t reserved_2; /* 11th - 12th are reserved bytes */
    uint16_t sample; /* 13th - 14th are sample period time */
} __attribute__ ((packed));

/* make a struct for the return from the reading command */
struct power_reading {
    uint8_t grp_id; /* first byte: Group Extension ID */
    uint16_t curr_pwr;
    uint16_t min_sample;
    uint16_t max_sample;
    uint16_t avg_pwr;
    uint32_t time_stamp; /* time since epoch */
    uint32_t sample;
    uint8_t state;
} __attribute__ ((packed));

/* make a struct for the return from the capabilites command */
struct capabilities {
    uint8_t grp_id; /* first byte: Group Extension ID */
    uint16_t conformance;
    uint8_t revision;
    uint8_t data_byte1;
    uint8_t data_byte2;
    uint8_t data_byte3;
    uint8_t data_byte4;
} __attribute__ ((packed));

/* make a struct for the return from the sensor info command */
struct sensor_info {
    uint8_t grp_id; /* first byte: Group Extension ID */
    uint8_t i_instances;
    uint8_t i_records;
    
} __attribute__ ((packed));

/* make a struct for the return from the get asset tag command */
struct asset_tag {
    uint8_t grp_id; /* first byte: Group Extension ID */
    uint8_t length;
    const char tag[16];
} __attribute__ ((packed));

int dcmi_main(struct ipmi_intf * intf, int argc, char ** argv);


#define BUILD_CAPABILITY_CMD(req) \
    memset(&req, 0, sizeof(req)); \
    req.msg.netfn = IPMI_NETFN_DCGRP; \
    req.msg.cmd = DCMI_CMD_CAPABILITY; \
    req.msg.data = msg_data; \
    req.msg.data_len = 2

#define BUILD_GET_ASSET_TAG_CMD(req) \
    memset(&req, 0, sizeof(req)); \
    req.msg.netfn = IPMI_NETFN_DCGRP; \
    req.msg.cmd = DCMI_CMD_GETASSET; \
    req.msg.data = msg_data; \
    req.msg.data_len = 3

#define BUILD_GET_SNSR_INFO_CMD(req) \
    memset(&req, 0, sizeof(req)); \
    req.msg.netfn = IPMI_NETFN_DCGRP; \
    req.msg.cmd = DCMI_CMD_GETSNSR; \
    req.msg.data = msg_data; \
    req.msg.data_len = 5
  
#define BUILD_GET_PWR_READING_CMD(req) \
    memset(&req, 0, sizeof(req)); \
    req.msg.netfn = IPMI_NETFN_DCGRP; \
    req.msg.cmd = DCMI_CMD_GETRED; \
    req.msg.data = msg_data; \
    req.msg.data_len = 4
    
#define BUILD_GET_PWR_LIMIT_CMD(req) \
    memset(&req, 0, sizeof(req)); \
    req.msg.netfn = IPMI_NETFN_DCGRP; \
    req.msg.cmd = DCMI_CMD_GETLMT; \
    req.msg.data = msg_data; \
    req.msg.data_len = 3
    
#define BUILD_SET_PWR_LIMIT_CMD(req) \
    memset(&req, 0, sizeof(req)); \
    req.msg.netfn = IPMI_NETFN_DCGRP; \
    req.msg.cmd = DCMI_CMD_SETLMT; \
    req.msg.data = msg_data; \
    req.msg.data_len = 15
    
#define BUILD_ACTIVATE_DEACTIVATE_PWR_LIMIT_CMD(req) \
    memset(&req, 0, sizeof(req)); \
    req.msg.netfn = IPMI_NETFN_DCGRP; \
    req.msg.cmd = DCMI_CMD_PWRACT; \
    req.msg.data = msg_data; \
    req.msg.data_len = 4
  
#endif /*IPMI_DCMI_H*/
