/*
 * Copyright (C) 2008 Intel Corporation.
 * All rights reserved
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, 
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR 
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, 
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, 
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. 
 *
 */

/*******************************************************************************
 * Theory of operation                                                         *
 *                                                                             *
 * DCMI is the Data Center Management Interface which defines the commands     *
 * suitable for data centers.                                                  *
 * DCMI incorporates the ability to locate a system with DCMI functionality,   *
 * its available temperature sensors, and power limiting control.              *
 *                                                                             *
 * All of the available DCMI commands are contained in a struct with a numeric *
 * value and a string.  When the user specifies a command the string is        *
 * compared to one of several structs and is then given a numeric value based  *
 * on the matched string.  A case statement is used to select the desired      *
 * action from the user.  If an invalid string is entered, or a string that is *
 * not a command option is entered, the available commands are printed to the  *
 * screen.  This allows the strings to be changed quickly with the DCMI spec.  *
 *                                                                             *
 * Each called function usually executes whichever command was requested to    *
 * keep the main() from being overly complicated.                              *
 *                                                                             *
 * This code conforms to the 1.0 DCMI Specification                            *
 *                                                                             *
 ******************************************************************************/

#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <math.h>
#include <unistd.h>
#include <sys/types.h>
#include <time.h>

#include <ipmitool/dcmi.h>

#include <ipmitool/helper.h>
#include <ipmitool/ipmi.h>
#include <ipmitool/log.h>
#include <ipmitool/ipmi_intf.h>
#include <ipmitool/ipmi_strings.h>

#include <ipmitool/ipmi_mc.h>
#include <ipmitool/ipmi_entity.h>
#include <ipmitool/ipmi_constants.h>
#include <ipmitool/ipmi_sensor.h>

extern int verbose;

/*******************************************************************************
 * The structs below are the DCMI command option strings.  They are printed    *
 * when the user does not issue enough options or the wrong ones.  The reason  *
 * that the DMCI command strings are in a struct is so that when the           *
 * specification changes, the strings can be changed quickly with out having   *
 * to change a lot of the code in the main().                                  *
 ******************************************************************************/

/* Main set of DCMI commands */
const struct dcmi_cmd dcmi_cmd_vals[] = {
    { 0x00, "discover", "  Used to discover supported DCMI capabilities" },
    { 0x01, "power", "     Platform power limit command options"         },
    { 0x02, "sensors", "   Prints the available DCMI sensors"            },
    { 0x03, "asset_tag", " Prints the platforms asset tag"               },
    { 0xFF, NULL, NULL                                                  }
};

/* get capabilites */
const struct dcmi_cmd dcmi_capable_vals[] = {
    { 0x01, "platform", "            Lists the system capabilities" },
    { 0x02, "mandatory_attributes", "Lists SEL, identification and"
        "temperature attributes"                                    },
    { 0x03, "optional_attributes", " Lists power capabilities"      },
    { 0x04, "managebility access", " Lists OOB channel information" },
    { 0xFF, NULL, NULL                                              }
};

/* platform capabilities                                                       
 * Since they are actually in two bytes, we need three structs to make this human
 * readable... */
const struct dcmi_cmd dcmi_mandatory_platform_capabilities[] = {
    { 0x01, "Identification support available", "" },
    { 0x02, "SEL logging available", ""            },
    { 0x03, "Chassis power available", ""          },
	{ 0x04, "Temperature monitor available", ""	 },
	{ 0xFF, NULL, NULL					 }
};

/* optional capabilities */
const struct dcmi_cmd dcmi_optional_platform_capabilities[] = {
    { 0x01, "Power management available", ""       },	
	{ 0xFF, NULL, NULL					 }
};

/* access capabilties */
const struct dcmi_cmd dcmi_management_access_capabilities[] = {
    { 0x01, "In-band KCS channel available", ""               },
    { 0x02, "Out-of-band serial TMODE available", ""          },
    { 0x03, "Out-of-band secondary LAN channel available", "" },
    { 0x04, "Out-of-band primary LAN channel available", ""   },
    { 0x05, "SOL enabled", ""                       },
    { 0x06, "VLAN capable", ""		},
    { 0xFF, NULL, NULL                              }
 };

/* identification capabilities */
const struct dcmi_cmd dcmi_id_capabilities_vals[] = {
	 { 0x01, "GUID", "" 		 },
	 { 0x02, "DHCP hostname", "" },
	 { 0x03, "Asset tag", ""	},
	 { 0xFF, NULL, NULL 		 }
};

/* temperature monitoring capabilities */
const struct dcmi_cmd dcmi_temp_monitoring_vals[] = {
	 { 0x01, "inlet", "    Inlet air temperature sensors"  },
	 { 0x02, "cpu", "	   CPU temperature sensors" 	   },
	 { 0x03, "baseboard", "Baseboard temperature sensors"  },
	 { 0xff, NULL, NULL 								   }
};

/* These are not comands.  These are the DCMI temp sensors and their numbers
 * If new sensors are added, they need to be added to this list with their
 * sensor number */
const struct dcmi_cmd dcmi_discvry_snsr_vals[] = {
    { 0x40, "Inlet", "    Inlet air temperature sensors"  },
    { 0x41, "CPU", "      CPU temperature sensors"        },
    { 0x42, "Baseboard", "Baseboard temperature sensors"  },
    { 0xff, NULL, NULL                                    }
};

/* power management/control commands */
const struct dcmi_cmd dcmi_pwrmgmt_vals[] = {
    { 0x00, "reading", "   Get power related readings from the system" },
    { 0x01, "get_limit", " Get the configured power limits"            },
    { 0x02, "set_limit", " Set a power limit option"                   },
    { 0x03, "activate", "  Activate the set power limit"               },
    { 0x04, "deactivate", "Deactivate the set power limit"             },
    { 0xFF, NULL, NULL                                                 }
};

/* set power limit commands */
const struct dcmi_cmd dcmi_pwrmgmt_set_usage_vals[] = {
    { 0x00, "action", "    <disabled | power_off>"                      },
    { 0x01, "limit", "     <number in Watts from 0 to 65535>"           },
    { 0x02, "correction", "<number in milliseconds from 6000 to 28751>" },
    { 0x03, "sample", "    <number in seconds from 3 to 65>"            },
    { 0xFF, NULL, NULL                                                  }
};

/* DCMI command specific completion code results per 1.0 spec
 * 80h - parameter not supported.
 * 81h - attempt to set the ‘set in progress’ value (in parameter #0) when not
 *       in the ‘set complete’ state. (This completion code provides a way to
 *       recognize that another party has already ‘claimed’ the parameters)
 * 82h - attempt to write read-only parameter
 * 82h - set not supported on selected channel (e.g. channel is session-less.)
 * 83h - access mode not supported
 * 84h – Power Limit out of range
 * 85h – Correction Time out of range
 * 89h – Statistics Reporting Period out of range
 */
const struct valstr dcmi_ccode_vals[] = {
    { 0x80, "Parameter not supported"                                },
    { 0x81, "Something else has already claimed these parameters"    },
    { 0x82, "Not supported or failed to write a read-only parameter" },
    { 0x83, "Access mode is not supported"                           },
    { 0x84, "Power limit out of range"                               },
    { 0x85, "Correction time out of range"                           },
    { 0x89, "Sample period out of range"                             },
    { 0xFF, NULL                                                     }
};

/* power management/set action commands */
const struct dcmi_cmd dcmi_pwrmgmt_action_vals[] = {
    { 0x00, "disabled", ""  },
    { 0x01, "power_off", "" },
    { 0xFF, NULL, NULL      }
};

/* End strings */

/*******************************************************************************
 * This was taken from print_valstr() from helper.c.  It serves the same       *
 * purpose but with out the extra formatting.  This function simply prints     *
 * the dcmi_cmd struct provided.  verthorz specifies to print vertically or    *
 * horizontally.  If the string is printed horizontally then a | will be       *
 * printed between each instance of vs[i].str until it is NULL                 *
 *                                                                             *
 * @vs:         value string list to print                                     *
 * @title:      name of this value string list                                 *
 * @loglevel:   what log level to print, -1 for stdout                         *
 * @verthorz:   printed vertically or horizontally, 0 or 1                     *
 ******************************************************************************/
void print_strs(const struct dcmi_cmd * vs, const char * title, int loglevel,
    int verthorz) {
    int i;

    if (vs == NULL)
        return;

    if (title != NULL) {
        if (loglevel < 0)
            printf("\n%s\n", title);
        else
            lprintf(loglevel, "\n%s", title);
    }
    for (i = 0; vs[i].str != NULL; i++) {
        if (loglevel < 0) {
            if (vs[i].val < 256)
                if (verthorz == 0)
                    printf("    %s    %s\n", vs[i].str, vs[i].desc);
                else
                    printf("%s", vs[i].str);
            else if (verthorz == 0)
                printf("    %s    %s\n", vs[i].str, vs[i].desc);
            else
                printf("%s", vs[i].str);
        } else {
            if (vs[i].val < 256)
                lprintf(loglevel, "    %s    %s", vs[i].str, vs[i].desc);
            else
                lprintf(loglevel, "    %s    %s", vs[i].str, vs[i].desc);
        }
        /* Check to see if this is NOT the last element in vs.str if true
         * print the | else don't print anything. */
        if ((verthorz == 1) && (vs[i+1].str != NULL))
            printf(" | ");
    }
    if (verthorz == 0)
        if (loglevel < 0)
            printf("\n");
        else
            lprintf(loglevel, "");
}

/*******************************************************************************
 * This was taken from str2val() from helper.c.  It serves the same            *
 * purpose but with the addition of a desc field from the structure.           *
 * This function converts the str from the dcmi_cmd struct provided to the     *
 * value associated to the compared string in the struct.                      *
 *                                                                             *
 * @str:        string to compare against                                      *
 * @vs:         dcmi_cmd structure                                             *
 ******************************************************************************/
uint16_t str2val2(const char *str, const struct dcmi_cmd *vs)
{
    int i;

    if (vs == NULL || str == NULL)
        return 0xff;
	
    for (i = 0; vs[i].str != NULL; i++) {
        if (strncasecmp(vs[i].str, str, __maxlen(str, vs[i].str)) == 0)
            return vs[i].val;
    }

    return vs[i].val;
}

/*******************************************************************************
 * This was taken from val2str() from helper.c.  It serves the same            *
 * purpose but with the addition of a desc field from the structure.           *
 * This function converts the val and returns a string from the dcmi_cmd       *
 * struct provided in the struct.                                              *
 *                                                                             *
 * @val:        value to compare against                                       *
 * @vs:         dcmi_cmd structure                                             *
 ******************************************************************************/
const char * val2str2(uint16_t val, const struct dcmi_cmd *vs)
{
    static char un_str[32];
    static char *ErrorMsg = "\nval2str2: Error, no valid command\n";
    int i;

    if (vs == NULL)
        return ErrorMsg;
	
    for (i = 0; vs[i].str != NULL; i++) {
        if (vs[i].val == val)
            return vs[i].str;
    }

    memset(un_str, 0, 32);
    snprintf(un_str, 32, "Unknown (0x%x)", val);

    return un_str;
}

/*******************************************************************************
 * check the DCMI response   from the BMC                                      *
 * @rsp:       Response data structure                                      *
 *                                                                             *
 ******************************************************************************/
static int chk_rsp(struct ipmi_rs * rsp) {
    
    /* if the response from the intf is NULL then the BMC is experiencing
     * some issue and cannot complete the command */
    if (rsp == NULL) {
        lprintf(LOG_ERR, "\n    Unable to get DCMI information");
        return 1;
    }
    
    /* if the completion code is greater than zero there was an error.  We'll
     * use val2str from helper.c to print the error from either the DCMI
     * completion code struct or the generic IPMI completion_code_vals struct
     */
    if ((rsp->ccode >= 0x80) && (rsp->ccode <= 0x8F)) {
        lprintf(LOG_ERR, "\n    DCMI request failed because: %s",
            val2str(rsp->ccode, dcmi_ccode_vals));
        return 1;
    } else if (rsp->ccode > 0) {
        lprintf(LOG_ERR, "\n    DCMI request failed because: %s",
            val2str(rsp->ccode, completion_code_vals));
        return 1;
    }
    
    /* check to make sure this is a DCMI firmware */
    if(rsp->data[0] != DCMI_GROUP) {
        printf("\n    A valid DCMI response was not returned!");
        return 1;
    }
    
    return 0;
}
/*******************************************************************************
 *  Get capabilities ipmi response                                             *
 *                                                                             *
 * This function returns the available capabilities of the platform.           *
 * The reason it returns in the rsp struct is so that it can be used for other *
 * purposes.                                                                   *
 *                                                                             *
 * returns ipmi response structure                                             *
 *                                                                             *
 * @intf:   ipmi interface handler                                             *
 * @selector: Parameter selector
 ******************************************************************************/
struct ipmi_rs * 
dcmi_getcapabilities(struct ipmi_intf * intf, uint8_t selector) {
    struct ipmi_rq req; /* request data to send to the BMC */
    uint8_t msg_data[2]; /* 'raw' data to be sent to the BMC */

    msg_data[0] = DCMI_GROUP; /* Group Extension Identification */
    msg_data[1] = selector;

    BUILD_CAPABILITY_CMD(req);

    return intf->sendrecv(intf, &req);
}
/* end capabilities struct */

/*******************************************************************************
/* Displays capabilities from structure
* returns void								  *
*																			  *
* @cmd:		   dcmi_cmd structure											  *
* @data_val:  holds value of what to display											  *
******************************************************************************/
void display_capabilities_attributes(
		const struct dcmi_cmd *cmd, uint8_t data_val)
{
	uint8_t i;
	for (i = 0x01; cmd[i-1].val != 0xFF; i++) {
		if (data_val & (1<<(i-1))) {
			printf("		%s\n", val2str2(i, cmd));
		}
	}
}

/*******************************************************************************
 * This is the get DCMI Capabilities function to see what the BMC supports.    *
 *                                                                             *
 * returns 0 with out error -1 with any errors                                 *
 *                                                                             *
 * @intf:      ipmi interface handler                                          *
 * @selector:  selection parameter                                             *
 ******************************************************************************/
static int
dcmi_prnt_getcapabilities(struct ipmi_intf * intf, uint8_t selector) {
    
    uint8_t i;
		uint8_t bit_shifter = 0;
    struct ipmi_rs * rsp = NULL;
    struct capabilities *cape = NULL;
    
    rsp = dcmi_getcapabilities(intf, selector);

    if(chk_rsp(rsp))
        return -1;

    /* if there were no errors, the command worked! */
    cape = (struct capabilities *)rsp->data;
    
    /* check to make sure that this is a 1.0 command */
    if (cape->conformance != DCMI_CMD_CONFORM) {
        printf("ERROR!  This command is not available on this platform\n");
        return -1;
    }
    
    /* check to make sure that this is a rev .01 */
    if (cape->revision != 0x01) {
        printf("ERROR!  This command is not compatible with this version\n");
        return -1;
    }
    
    /* 0x01 - platform capabilities
     * 0x02 - Manageability Access Capabilities
     * 0x03 - SEL Capability
     * 0x04 - Identification Capability
     * 0x05 - LAN Out-Of-Band Capability
     * 0x06 - Serial Out-Of-Band TMODE Capability
     */
    switch (selector) {
        case 0x01:
            printf("    Supported DCMI capabilities:\n");
            /* loop through each of the entries in the first byte from the
             * struct */
            printf("\n         Mandatory platform capabilties\n");
			display_capabilities_attributes(
				dcmi_mandatory_platform_capabilities, cape->data_byte1);
            /* loop through each of the entries in the second byte from the
             * struct */           
            printf("\n         Optional platform capabilties\n");
			display_capabilities_attributes(
				dcmi_optional_platform_capabilities, cape->data_byte2);
            /* loop through each of the entries in the third byte from the
             * struct */           
            printf("\n         Managebility access capabilties\n");
			display_capabilities_attributes(
				dcmi_management_access_capabilities, cape->data_byte3);
            break;
        case 0x02:
            printf("\n    Mandatory platform attributes:\n");
            /* byte 1 & 2 data */
            printf("\n         SEL Attributes: ");
            printf("\n               SEL automatic rollover is ");
            /* mask the 2nd byte of the data response with 10000000b or 0x80   *
             * because of the endian-ness the 15th bit is in the second byte  */
            if ((cape->data_byte2 & 0x80))
                printf("enabled");
            else
                printf("not present");
            
            /* since the number of SEL entries is split across the two data
             * bytes we will need to bit shift and append them together again */
            /* cast cape.data_byte1 as 16 bits */
            uint16_t sel_entries = (uint16_t)cape->data_byte1;
            /* or sel_entries with byte 2 and shift it 8 places  */
            sel_entries |= (uint16_t)cape->data_byte2 << 8;
            printf("\n               %d SEL entries\n", sel_entries & 0xFFF);
             /* byte 3 data */
            printf("\n         Identification Attributes: \n");
			display_capabilities_attributes(
				 dcmi_id_capabilities_vals, cape->data_byte3);
             /* byte 4 data */
            printf("\n         Temperature Monitoring Attributes: \n");
			 display_capabilities_attributes(
				  dcmi_temp_monitoring_vals, cape->data_byte4);		
            break;
        case 0x03:
            printf("\n    Optional Platform Attributes: \n");
            /* Power Management */
            printf("\n         Power Management:\n");
            if (cape->data_byte1 == 0x40)
            {
                printf("                Slave address of device: 20h (BMC)\n" );
            }
            else
            {
                printf("                Slave address of device: %xh "
                "(Satellite/External controller)\n", cape->data_byte1 >> 1);
            }
			// Controller channel number (4-7) bits
            if ((cape->data_byte2>>4) == 0x00)
            {
                printf("                Channel number is 0h (Primary BMC)\n");
            }
			else
			{
                printf("                Channel number is %xh \n",(cape->data_byte2>>4));				
			}	
			// Device revision (0-3)
            printf("                    Device revision is %d \n", cape->data_byte2 &0xf);
            break;
        case 0x04:
            /* LAN */
            printf("\n    Manageability Access Attributes: \n");
            if (cape->data_byte1 == 0xFF)
                printf("         Primary LAN channel is not available for OOB\n");
            else
                printf("         Primary LAN channel number: %d is available\n",
                    cape->data_byte1);
            
            if (cape->data_byte2 == 0xFF)
                printf("         Secondary LAN channel is not available for OOB\n");
            else
                printf("         Secondary LAN channel number: %d is available\n",
                    cape->data_byte2);
            /* serial */
            if (cape->data_byte3 == 0xFF)
                printf("         No serial channel is available\n");

            else
                printf("         Serial channel number: %d is available\n",
                    cape->data_byte3);
            break;
        default:
            return -1;
    }

    return 0;
    /* return intf->sendrecv(intf, &req); */
}

/*******************************************************************************
 * This is the get asset tag command.  This checks the length of the asset tag *
 * with the first read, then reads n number of bytes thereafter to get the     *
 * complete asset tag.                                                         *
 *                                                                             *
 * @intf:   ipmi interface handler                                             *
 * @offset: where to start reading the asset tag                               *
 * @length: how much to read                                                   *
 *                                                                             *
 * returns ipmi_rs structure                                                   *
 ******************************************************************************/
struct ipmi_rs * dcmi_getassettag(
    struct ipmi_intf * intf, uint8_t offset, uint8_t length) {
    
    struct ipmi_rq req; /* request data to send to the BMC */
    uint8_t msg_data[3]; /* 'raw' data to be sent to the BMC */
        
    msg_data[0] = DCMI_GROUP; /* Group Extension Identification */
    msg_data[1] = offset; /* offset 0 */
    msg_data[2] = length; /* read one byte */

		BUILD_GET_ASSET_TAG_CMD(req);

    return intf->sendrecv(intf, &req);
}

/*******************************************************************************
 * This is the get asset tag command.  The function first checks to see if the *
 * platform is capable of getting the asset tag by calling the getcapabilities *
 * function and checking the response.  Then it checks the length of the asset *
 * tag with the first read, then x number of reads thereafter to get the asset *
 * complete asset tag then print it.                                           *
 *                                                                             *
 * @intf:   ipmi interface handler                                             *
 *                                                                             *
 * returns 0 if no failure, -1 with a failure                                  *
 ******************************************************************************/
static int dcmi_prnt_getassettag(struct ipmi_intf * intf) {
    uint8_t data_byte2;
    struct ipmi_rs * rsp = NULL; /* ipmi response */
	uint8_t taglength = 0;
	uint8_t getlength = 0;
	uint8_t offset = 0;
	uint8_t i;
	
    /* now let's get the asset tag length */
    rsp = dcmi_getassettag(intf, 0, 0);

    if(chk_rsp(rsp))
        return -1;
    
    taglength = rsp->data[1];
        
    printf("\n    Asset tag: ");
    while (taglength) {
		getlength = taglength/0x10 ? 0x10 : taglength%0x10;
        rsp = dcmi_getassettag(intf, offset, getlength);

        if(chk_rsp(rsp))
            return -1;
 
        for (i=0; i<getlength; i++) {
            printf("%c", rsp->data[i+2]);
        }
		offset += getlength;
        taglength -= getlength;
    }
   
    printf("\n");   
    return 0;
}

/*******************************************************************************
 * Issues a discovery command to see what sensors are available on the target. *
 * system.  
 *                                                                             *
 * @intf:   ipmi interface handler                                             *
 * @isnsr:  entity ID                                                          *
 * @offset:   offset (Entity instace start)                                                   *
 *                                                                             *
 * returns ipmi_rs structure                                                   *
 ******************************************************************************/
struct ipmi_rs * dcmi_discvry_snsr( 
	struct ipmi_intf * intf, uint8_t isnsr, uint8_t offset) {
    
    struct ipmi_rq req; /* ipmi request struct */
    uint8_t msg_data[5]; /* number of request data bytes */

    msg_data[0] = DCMI_GROUP; /* Group Extension Identification */
    msg_data[1] = 0x01; /* Senser Type = Temp (01h) */
    msg_data[2] = isnsr; /* Sensor Number */
    msg_data[3] = 0x00; /* Entity Instance, set to read all instances */
    msg_data[4] = offset; /* Entity instance start */

    BUILD_GET_SNSR_INFO_CMD(req);

    return intf->sendrecv(intf, &req);
}

static int dcmi_print_sensor_info(struct ipmi_intf *intf, uint16_t rec_id);
/*******************************************************************************
 * DCMI sensor discovery                                                       *
 * Uses the dcmi_discvry_snsr_vals struct to print its string and              *
 * uses the numeric values to request the sensor sdr record id.                *
 *                                                                             *
 * @intf:   ipmi interface handler                                             *
 * @isnsr:  entity ID                                                          *
 * @ient:   sensor entity id                                                   *
 *                                                                             *
 ******************************************************************************/
static int dcmi_prnt_discvry_snsr(struct ipmi_intf * intf, uint8_t isnsr) {
  int i = 0;
  struct ipmi_rs * rsp = NULL; /* ipmi response */
  uint8_t records = 0;
  int8_t instances = 0;
	uint8_t offset = 0;
	uint16_t record_id = 0;
	uint8_t id_buff[16]; /* enough for 8 record IDs (in one response) based on DCMI 1.0 */
	
	
	/* query the BMC with the entity ID and find out the number of instances */
  /* buffer for rsp is statically defined in ipmi_lan_recv_packet, don't free it */
	rsp = dcmi_discvry_snsr(intf, isnsr, 0); 

	/* the response code is stored in rsp->ccode and chk_rsp() checks it */
  if(chk_rsp(rsp))
      return -1;

	/* Get the total number of instances for this Entity ID, could be greater than 8 */
	instances = rsp->data[1];
	
	printf("\n%s: %d temperature sensor%s found:\n",
		val2str2(isnsr, dcmi_discvry_snsr_vals), 
		instances, 
		(instances > 1) ? "s" : "");

  /* instances is the number of SDRs with the same Entity ID */			
	while(instances > 0)
	{
		/* number of Record ID in this response, must not be greater than 8 (see DCMI 1.0) */
		records = rsp->data[2];
		if(records > 8)
			return -1;
		
		/* save the record ID list since rsp->data[] will be overwritten by subsequent ipmi_xxx calls */
		memcpy(id_buff, &rsp->data[3], records * 2);	  	
		
		/* print the sensor info for each record ID */
		for (i=0; i<records; i++) {
		 /* Record ID is in little endian format, conver it to unsigned short format */
		  record_id = (id_buff[2*i + 1] << 8) + id_buff[2*i];
			printf("Record ID 0x%04x: ", record_id);
			dcmi_print_sensor_info(intf, record_id);
		}
		
		/* adjust the offset in case there are more than 8 record IDs */
		offset += records;
		instances -= records;
		
		if(instances)
		{
			rsp = dcmi_discvry_snsr(intf, isnsr, offset);
			if(chk_rsp(rsp))
				return -1;
		}
	}
	
  return 0;
}

/* end sensor discovery */

/*******************************************************************************
 *  Power Management get power reading                                         *
 *                                                                             *
 * @intf:   ipmi interface handler                                             *
 ******************************************************************************/
static int dcmi_pwr_rd(struct ipmi_intf * intf) {
    struct ipmi_rs * rsp = NULL;
    struct ipmi_rq req;
    uint8_t msg_data[4]; /* number of request data bytes */
    struct power_reading *val = NULL;

    msg_data[0] = DCMI_GROUP; /* Group Extension Identification */
    msg_data[1] = 0x01; /* Mode Power Status */
    msg_data[2] = 0x00; /* reserved */
    msg_data[3] = 0x00; /* reserved */

		BUILD_GET_PWR_READING_CMD(req);

    rsp = intf->sendrecv(intf, &req);

    if(chk_rsp(rsp))
        return -1;
    
    /* rsp->data[0] is equal to response data byte 2 in spec */
    val = (struct power_reading *)rsp->data;
    printf("\n");
    printf("    Instantaneous power reading:              %d Watts\n",
        val->curr_pwr);
    printf("    Minimum Sampling period:                  %d Seconds\n",
        val->min_sample);
    printf("    Maximum Sampling period:                  %d Seconds\n",
        val->max_sample);
    printf("    Average power reading over sample period: %d Watts\n",
        val->avg_pwr);
    printf("    IPMI Timestamp:                           %s",
        ctime((time_t*)&val->time_stamp));
    printf("    Sampling period:                          %d Milliseconds\n",
        val->sample);
    printf("    Power reading state is:                   ");
    /* mask the rsp->data so that we only care about bit 6 */
    if((val->state & 0x40) == 0x40) {
        printf("activated");
    } else {
        printf("deactivated");
    }
    printf("\n\n");
    return 0;
}
/* end Power Management get reading */

/******************************************************************************
 *  Power Management get limit ipmi response
 *
 * This function returns the currently set power management settings as an
 * ipmi response structure.  The reason it returns in the rsp struct is so
 * that it can be used in the set limit [slimit()] function to populate
 * un-changed or un-edited values.
 *
 * returns ipmi response structure
 *                                                                             *
 * @intf:   ipmi interface handler                                             *
 ******************************************************************************/
struct ipmi_rs * dcmi_pwr_glimit(struct ipmi_intf * intf) {
    struct ipmi_rq req;
    uint8_t msg_data[3]; /* number of request data bytes */

    msg_data[0] = DCMI_GROUP; /* Group Extension Identification */
    msg_data[1] = 0x00; /* reserved */
    msg_data[2] = 0x00; /* reserved */

    BUILD_GET_PWR_LIMIT_CMD(req);

    return intf->sendrecv(intf, &req);
}
/* end Power Management get limit response */

/*******************************************************************************
 *  Power Management print the get limit command
 *
 * This function calls the get limit function that returns an ipmi response.
 *
 * returns 0 else 1 with error
 * @intf:   ipmi interface handler                                             *
 ******************************************************************************/
static int dcmi_pwr_prnt_glimit(struct ipmi_intf * intf) {

    struct ipmi_rs * rsp = NULL;
    struct power_limit *val = NULL;

    rsp = dcmi_pwr_glimit(intf);

    if(chk_rsp(rsp))
        return -1;
    
    /* rsp->data[0] is equal to response data byte 2 in spec */
    /* printf("Group Extension Identification: %02x\n", rsp->data[0]); */
    val = (struct power_limit *)rsp->data;
       
    printf("\n    Exception actions: %s\n",
        val2str2(val->action, dcmi_pwrmgmt_action_vals));
    printf("    Power Limit:       %i Watts\n", val->limit);
    printf("    Correction time:   %i milliseconds\n", val->correction);
    printf("    Sampling period:   %i seconds\n", val->sample);
    printf("\n");
    return 0;
}
/* end print get limit */

/*******************************************************************************
 *  Power Management set limit
 *
 * Undocumented bounds:
 * Power limit: 0 - 0xFFFF
 * Correction period 5750ms to 28751ms or 0x1676 to 0x704F
 * sample period: 3 sec to 65 sec and 69+
 *
 * @intf:    ipmi interface handler                                            *
 * @option:  Power option to change                                            *
 * @value:   Value of the desired change                                       *
 ******************************************************************************/
static int dcmi_pwr_slimit(struct ipmi_intf * intf, const char * option,
    const char * value) {
    struct ipmi_rs * rsp = NULL; /* ipmi response */
    struct power_limit *val = NULL;
    struct ipmi_rq req; /* ipmi request (to send) */
    uint8_t msg_data[15]; /* number of request data bytes */
    uint32_t lvalue = strtol(value, NULL, 10);
    int i;
 
    
    rsp = dcmi_pwr_glimit(intf); /* get the power limit settings */

    if(chk_rsp(rsp))
        return -1;
    
    val = (struct power_limit *)rsp->data;
    
    /* same as above; sets the values of the val struct
    * DCMI group ID *
    * val.grp_id = rsp->data[0];
    * exception action *
    * val.action = rsp->data[3]; * 
    * 
    * power limit in Watts *
    * store 16 bits of the rsp from the 4th entity *
    * val.limit = *(uint16_t*)(&rsp->data[4]);
    * correction period in mS *
    * store 32 bits of the rsp from the 6th entity *
    * val.correction = *(uint32_t*)(&rsp->data[6]);
    * store 16 bits of the rsp from the 12th entity *
    * sample period in seconds *
    * val.sample = *(uint16_t*)(&rsp->data[12]);
    */

    switch (str2val2(option, dcmi_pwrmgmt_set_usage_vals)) {
    case 0x00:
        /* action */
        switch (str2val2(value, dcmi_pwrmgmt_action_vals)) {
        case 0x00:
            /* disable */
            val->action = 0;
            break;
        case 0x01:
            /* power_off */
            val->action = 1;
            break;
        case 0xFF:
            /* error - not a string we knew what to do with */
            return -1;
        }
        break;
    case 0x01:
        /* limit */
        val->limit = *(uint16_t*)(&lvalue);
        break;
    case 0x02:
        /* correction */
        val->correction = *(uint32_t*)(&lvalue);
        break;
    case 0x03:
        /* sample */
        val->sample = *(uint16_t*)(&lvalue);
        break;
    default:
        /* no valid options */
        return -1;
    }

    msg_data[0] = val->grp_id; /* Group Extension Identification */

    msg_data[1] = 0x00; /* reserved */
    msg_data[2] = 0x00; /* reserved */
    msg_data[3] = 0x00; /* reserved */

    msg_data[4] = val->action; /* exception action; 0x00 disables it */

    /* fill msg_data[5] with the first 16 bits of val.limit */
    *(uint16_t*)(&msg_data[5]) = val->limit;
    // msg_data[5] = 0xFF;
    // msg_data[6] = 0xFF;


    /* fill msg_data[7] with the first 32 bits of val.correction */
    *(uint32_t*)(&msg_data[7]) = val->correction;
    // msg_data[7] = 0x76;
    // msg_data[8] = 0x16;
    // msg_data[9] = 0x00;
    // msg_data[10] = 0x00;


    msg_data[11] = 0x00; /* reserved */
    msg_data[12] = 0x00; /* reserved */

    /* fill msg_data[7] with the first 16 bits of val.sample */
    *(uint16_t*)(&msg_data[13]) = val->sample;
    // msg_data[13] = 0x03;


//    memset(&req, 0, sizeof(req));
//    req.msg.netfn = IPMI_NETFN_DCGRP;
//    req.msg.cmd = DCMI_CMD_SETLMT;
//    req.msg.data = msg_data;
//    req.msg.data_len = 15;
    BUILD_SET_PWR_LIMIT_CMD(req);

    rsp = intf->sendrecv(intf, &req);

    if(chk_rsp(rsp))
        return -1;
    
    return 0;
}
/* end Power Management set limit */

/*******************************************************************************
 *  Power Management activate deactivate
 *
 * @intf:    ipmi interface handler                                            *
 * @option:  uint8_t - 0 to deactivate or 1 to activate                        *
 ******************************************************************************/
static int dcmi_pwr_actdeact(struct ipmi_intf * intf, uint8_t option) {
    struct ipmi_rs * rsp = NULL;
    struct ipmi_rq req;
    uint8_t msg_data[4]; /* number of request data bytes */

    msg_data[0] = DCMI_GROUP; /* Group Extension Identification */
    msg_data[1] = option; /* 0 = Deactivate 1 = Activate */
    msg_data[2] = 0x00; /* reserved */
    msg_data[3] = 0x00; /* reserved */

    BUILD_ACTIVATE_DEACTIVATE_PWR_LIMIT_CMD(req);

    rsp = intf->sendrecv(intf, &req);

    if(chk_rsp(rsp))
        return -1;
    
    printf("\n    Power limit successfully ");
    if (option == 0x00) {
        printf("deactivated");
    } else {
        printf("activated");
    }
    printf("\n");
    
    return 0;
}
/* end power management activate/deactivate */

/*******************************************************************************
 *  main
 *
 * @intf:   ipmi interface handler
 * @argc:   argument count
 * @argv:   argument vector
 ******************************************************************************/
int dcmi_main(struct ipmi_intf * intf, int argc, char **argv) {
    int rc = 0; /* Return code used when calling some of the dcmi commands */
    uint8_t ctl = 0;
    int i, ii, instances;
    struct ipmi_rs *rsp = NULL;
   
    if ((argc == 0) || (strncmp(argv[0], "help", 4) == 0)) {
        print_strs(
            dcmi_cmd_vals, "Data Center Management Interface commands", -1, 0);
        return -1;
    }

    /* start the cmd requested */
    switch (str2val2(argv[0], dcmi_cmd_vals)) {
    case 0x00:
        /* discover capabilities*/
        
        for (i = 1; dcmi_capable_vals[i-1].str != NULL; i++) {
            if((rc=dcmi_prnt_getcapabilities(intf, i)) < 0) {
                printf("Error discovering %s capabilities!\n",
                    val2str2(i, dcmi_capable_vals));
                return rc;
            }   
        }
        
        break;
        /*  end discover */

    case 0x01:
        /* power */

        argv++;
        if (argv[0] == NULL) {
            print_strs(dcmi_pwrmgmt_vals, "power <command>", -1, 0);
            return -1;
        }
        /* power management */
        switch (str2val2(argv[0], dcmi_pwrmgmt_vals)) {
        case 0x00:
            /* get reading */
            rc = dcmi_pwr_rd(intf);
            break;
        case 0x01:
            /* get limit */
            /* because the get limit function is also used to populate
             * unchanged values for the set limit command it returns an ipmi
             * response structure
             */
            rc = dcmi_pwr_prnt_glimit(intf);
            break;
        case 0x02:
            /* set limit */
            if (argc < 4) {
                print_strs(dcmi_pwrmgmt_set_usage_vals,
                    "set_limit <parameter> <value>", -1, 0);
                return -1;
            }
            
            /* loop through each parameter and value until we have neither */
            while ((argv[1] != NULL) && (argv[2] != NULL)) {
                rc = dcmi_pwr_slimit(intf, argv[1], argv[2]);
                /* catch any error that the set limit function returned */
                if (rc > 0) {
                    print_strs(dcmi_pwrmgmt_set_usage_vals,
                        "set_limit <parameter> <value>", -1, 0);
                    return -1;
                }
                /* the first argument is the command and the second is the 
                 * value.  Move argv two places; what is now 3 will be 1 */
                argv+=2;
            }
            rc = dcmi_pwr_prnt_glimit(intf);
            break;
        case 0x03:
            /* activate */
            rc = dcmi_pwr_actdeact(intf, 1);
            break;
        case 0x04:
            /* deactivate */
            rc = dcmi_pwr_actdeact(intf, 0);
            break;
        default:
            /* no valid options */
            print_strs(dcmi_pwrmgmt_vals, "power <command>", -1, 0);
            break;
        }
        /* power mgmt end */
        break;
        /* end power command */
    case 0x02:
        /* sensor print */
        /* Look for each item in the dcmi_discvry_snsr_vals struct
         * and if it exists, print the sdr record id(s) for it.
         * Use the val from each one as the sensor number. */
        for (i = 0; dcmi_discvry_snsr_vals[i].str != NULL; i++) {
            /* get all of the information about this sensor */
            rc = dcmi_prnt_discvry_snsr(
                intf, dcmi_discvry_snsr_vals[i].val);
                
            if(rc < 0)
            	return rc;
        }
        break;
        /* end sensor print */
    case 0x03:
        /* asset tag */
        rc = dcmi_prnt_getassettag(intf);
        if(rc < 0) {
            printf("Error getting asset tag!\n");
            return rc;
        }   
        break;
        /* end asset tag */
    default:
        /* couldn't detect what the user entered */
        print_strs(
            dcmi_cmd_vals, "Data Center Management Interface commands", -1, 0);
        return -1;
        break;
    }
    printf("\n");
    return rc;
}



/*******************************************************************************
 * Display DCMI sensor information                                             *
 * Uses the ipmi_sdr_get_next_header to read SDR header and compare to the     *
 * target Record ID. Then either ipmi_sensor_print_full or                     *
 * ipmi_sensor_print_compact is called to print the data                       *
 *                                                                             *
 * @intf:   ipmi interface handler                                             *
 * @rec_id: target Record ID                                                   *
 *                                                                             *
 ******************************************************************************/
static int dcmi_print_sensor_info(struct ipmi_intf *intf, uint16_t rec_id)
{
	struct sdr_get_rs *header;
	struct ipmi_sdr_iterator itr;
	int rc = 0;
	uint8_t *rec = NULL;

	itr.reservation = 0; 	/* only needed for partial SDR read with non-zero offset */
	itr.use_built_in = 0; /* netfunc = IPMI_NETFN_STORAGE */
	itr.next = rec_id;    /* this is the record we need, used by ipmi_sdr_get_next_header */

	rc = ipmi_sdr_get_reservation(intf, 0, &itr.reservation);  	/* step1: get the reservation ID */
	if(rc < 0 ) {
  	lprintf(LOG_DEBUG, "Cannot get reservation ID");
		return rc;
	}
		
	header = ipmi_sdr_get_next_header(intf, &itr);							/* step2: get the header to figure out the size of the SDR and type */
	if(NULL == header) {
		lprintf(LOG_DEBUG, "Header not found");
		return -1;
	}

	rec = ipmi_sdr_get_record(intf, header, &itr);							/* step3: read the SDR body */
	if(NULL == rec) {
		lprintf(LOG_DEBUG, "SDR not found");
		return -1;
	}

  /* now we can print the sensor value */
	if(header->type == SDR_RECORD_TYPE_FULL_SENSOR) {
		rc = ipmi_sensor_print_full(intf, (struct sdr_record_full_sensor *)rec);
	}
	else if(header->type == SDR_RECORD_TYPE_COMPACT_SENSOR) {
		rc = ipmi_sensor_print_compact(intf, (struct sdr_record_compact_sensor *)rec);
	}
	else
		rc = -1;

  /* free the memory allocated in ipmi_sdr_get_record */
	free(rec);	
	return rc;
}
