/*
 * Copyright (c) 2003 Sun Microsystems, Inc.  All Rights Reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * Redistribution of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 * 
 * Redistribution in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 * 
 * Neither the name of Sun Microsystems, Inc. or the names of
 * contributors may be used to endorse or promote products derived
 * from this software without specific prior written permission.
 * 
 * This software is provided "AS IS," without a warranty of any kind.
 * ALL EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND WARRANTIES,
 * INCLUDING ANY IMPLIED WARRANTY OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE OR NON-INFRINGEMENT, ARE HEREBY EXCLUDED.
 * SUN MICROSYSTEMS, INC. ("SUN") AND ITS LICENSORS SHALL NOT BE LIABLE
 * FOR ANY DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING
 * OR DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.  IN NO EVENT WILL
 * SUN OR ITS LICENSORS BE LIABLE FOR ANY LOST REVENUE, PROFIT OR DATA,
 * OR FOR DIRECT, INDIRECT, SPECIAL, CONSEQUENTIAL, INCIDENTAL OR
 * PUNITIVE DAMAGES, HOWEVER CAUSED AND REGARDLESS OF THE THEORY OF
 * LIABILITY, ARISING OUT OF THE USE OF OR INABILITY TO USE THIS SOFTWARE,
 * EVEN IF SUN HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGES.
 */

#include <stdlib.h>
#include <string.h>
#include <stdio.h>

#include <ipmitool/helper.h>
#include <ipmitool/ipmi.h>
#include <ipmitool/log.h>
#include <ipmitool/ipmi_intf.h>
#include <ipmitool/ipmi_strings.h>
#include <ipmitool/ipmi_chassis.h>



int
ipmi_chassis_power_status(struct ipmi_intf * intf)
{
   struct ipmi_rs * rsp;
   struct ipmi_rq req;

   memset(&req, 0, sizeof(req));
   req.msg.netfn = IPMI_NETFN_CHASSIS;
   req.msg.cmd = 0x1;
   req.msg.data_len = 0;

   rsp = intf->sendrecv(intf, &req);
   if (rsp == NULL) {
      lprintf(LOG_ERR, "Unable to get Chassis Power Status");
      return -1;
   }
   if (rsp->ccode > 0) {
      lprintf(LOG_ERR, "Get Chassis Power Status failed: %s",
         val2str(rsp->ccode, completion_code_vals));
      return -1;
   }

   return rsp->data[0] & 1;
}

static int
ipmi_chassis_print_power_status(struct ipmi_intf * intf)
{
   int ps = ipmi_chassis_power_status(intf);

   if (ps < 0)
      return -1;

   printf("Chassis Power is %s\n", ps ? "on" : "off");

   return 0;
}

int
ipmi_chassis_power_control(struct ipmi_intf * intf, uint8_t ctl)
{
   struct ipmi_rs * rsp;
   struct ipmi_rq req;

   memset(&req, 0, sizeof(req));
   req.msg.netfn = IPMI_NETFN_CHASSIS;
   req.msg.cmd = 0x2;
   req.msg.data = &ctl;
   req.msg.data_len = 1;

   rsp = intf->sendrecv(intf, &req);
   if (rsp == NULL) {
      lprintf(LOG_ERR, "Unable to set Chassis Power Control to %s",
         val2str(ctl, ipmi_chassis_power_control_vals));
      return -1;
   }
   if (rsp->ccode > 0) {
      lprintf(LOG_ERR, "Set Chassis Power Control to %s failed: %s",
         val2str(ctl, ipmi_chassis_power_control_vals),
         val2str(rsp->ccode, completion_code_vals));
      return -1;
   }

   printf("Chassis Power Control: %s\n",
          val2str(ctl, ipmi_chassis_power_control_vals));

#if 0	/* this can cause sessions to hang around after power commands */
   /* sessions often get lost when changing chassis power */
   intf->abort = 1;
#endif

   return 0;
}

static int
ipmi_chassis_identify(struct ipmi_intf * intf, char * arg)
{
   struct ipmi_rq req;
   struct ipmi_rs * rsp;

   struct {
      uint8_t interval;
      uint8_t force_on;
   } identify_data;

   memset(&req, 0, sizeof(req));
   req.msg.netfn = IPMI_NETFN_CHASSIS;
   req.msg.cmd = 0x4;

   if (arg != NULL) {
      if (strncmp(arg, "force", 5) == 0) {
         identify_data.interval = 0;
         identify_data.force_on = 1;
      } else {
         identify_data.interval = (uint8_t)atoi(arg);
         identify_data.force_on = 0;
      }
      req.msg.data = (uint8_t *)&identify_data;
      /* The Force Identify On byte is optional and not
       * supported by all devices-- if force is not specified,
       * we pass only one data byte; if specified, we pass two
       * data bytes and check for an error completion code
       */	
      req.msg.data_len = (identify_data.force_on) ? 2 : 1;
   }

   rsp = intf->sendrecv(intf, &req);
   if (rsp == NULL) {
      lprintf(LOG_ERR, "Unable to set Chassis Identify");
      return -1;
   }
   if (rsp->ccode > 0) {
      lprintf(LOG_ERR, "Set Chassis Identify failed: %s",
         val2str(rsp->ccode, completion_code_vals));
      if (identify_data.force_on != 0) {
         /* Intel SE7501WV2 F/W 1.2 returns CC 0xC7, but
          * the IPMI v1.5 spec does not standardize a CC
          * if unsupported, so we warn
          */
         lprintf(LOG_WARNING, "Chassis may not support Force Identify On\n");
      }
      return -1;
   }

   printf("Chassis identify interval: ");
   if (arg == NULL) {
      printf("default (15 seconds)\n");
   } else {
      if (identify_data.force_on != 0) {
         printf("indefinite\n");
      } else {
         if (identify_data.interval == 0)
            printf("off\n");
         else
            printf("%i seconds\n", identify_data.interval);
      }
   }
   return 0;
}



int
ipmi_chassis_status(struct ipmi_intf * intf)
{
   struct ipmi_rs * rsp;
   struct ipmi_rq req;

   memset(&req, 0, sizeof(req));
   req.msg.netfn = IPMI_NETFN_CHASSIS;
   req.msg.cmd = 0x1;

   rsp = intf->sendrecv(intf, &req);
   if (rsp == NULL) {
      lprintf(LOG_ERR, "Error sending Chassis Status command");
      return -1;
   }
   if (rsp->ccode > 0) {
      lprintf(LOG_ERR, "Error sending Chassis Status command: %s",
         val2str(rsp->ccode, completion_code_vals));
      return -1;
   }

   /* byte 1 */
   printf("System Power         : %s\n", (rsp->data[0] & 0x1) ? "on" : "off");
   printf("Power Overload       : %s\n", (rsp->data[0] & 0x2) ? "true" : "false");
   printf("Power Interlock      : %s\n", (rsp->data[0] & 0x4) ? "active" : "inactive");
   printf("Main Power Fault     : %s\n", (rsp->data[0] & 0x8) ? "true" : "false");
   printf("Power Control Fault  : %s\n", (rsp->data[0] & 0x10) ? "true" : "false");
   printf("Power Restore Policy : ");
   switch ((rsp->data[0] & 0x60) >> 5) {
   case 0x0:
      printf("always-off\n");
      break;
   case 0x1:
      printf("previous\n");
      break;
   case 0x2:
      printf("always-on\n");
      break;
   case 0x3:
   default:
      printf("unknown\n");
   }

   /* byte 2 */
   printf("Last Power Event     : ");
   if (rsp->data[1] & 0x1)
      printf("ac-failed ");
   if (rsp->data[1] & 0x2)
      printf("overload ");
   if (rsp->data[1] & 0x4)
      printf("interlock ");
   if (rsp->data[1] & 0x8)
      printf("fault ");
   if (rsp->data[1] & 0x10)
      printf("command");
   printf("\n");

   /* byte 3 */
   printf("Chassis Intrusion    : %s\n", (rsp->data[2] & 0x1) ? "active" : "inactive");
   printf("Front-Panel Lockout  : %s\n", (rsp->data[2] & 0x2) ? "active" : "inactive");
   printf("Drive Fault          : %s\n", (rsp->data[2] & 0x4) ? "true" : "false");
   printf("Cooling/Fan Fault    : %s\n", (rsp->data[2] & 0x8) ? "true" : "false");

        if (rsp->data_len > 3) {
      /* optional byte 4 */
      if (rsp->data[3] == 0) {
         printf("Front Panel Control  : none\n");
      } else {
         printf("Sleep Button Disable : %s\n", (rsp->data[3] & 0x80) ? "allowed" : "not allowed");
         printf("Diag Button Disable  : %s\n", (rsp->data[3] & 0x40) ? "allowed" : "not allowed");
         printf("Reset Button Disable : %s\n", (rsp->data[3] & 0x20) ? "allowed" : "not allowed");
         printf("Power Button Disable : %s\n", (rsp->data[3] & 0x10) ? "allowed" : "not allowed");
         printf("Sleep Button Disabled: %s\n", (rsp->data[3] & 0x80) ? "true" : "false");
         printf("Diag Button Disabled : %s\n", (rsp->data[3] & 0x40) ? "true" : "false");
         printf("Reset Button Disabled: %s\n", (rsp->data[3] & 0x20) ? "true" : "false");
         printf("Power Button Disabled: %s\n", (rsp->data[3] & 0x10) ? "true" : "false");
      }
        }

   return 0;
}




int
ipmi_power_main(struct ipmi_intf * intf, int argc, char ** argv)
{
   int rc = 0;
   uint8_t ctl = 0;

   if ((argc < 1) || (strncmp(argv[0], "help", 4) == 0)) {
      lprintf(LOG_NOTICE, "chassis power Commands: status, on, off, cycle, reset, diag, soft");
      return 0;
   }
   if (strncmp(argv[0], "status", 6) == 0) {
      rc = ipmi_chassis_print_power_status(intf);
      return rc;
   }
   if ((strncmp(argv[0], "up", 2) == 0) || (strncmp(argv[0], "on", 2) == 0))
      ctl = IPMI_CHASSIS_CTL_POWER_UP;
   else if ((strncmp(argv[0], "down", 4) == 0) || (strncmp(argv[0], "off", 3) == 0))
      ctl = IPMI_CHASSIS_CTL_POWER_DOWN;
   else if (strncmp(argv[0], "cycle", 5) == 0)
      ctl = IPMI_CHASSIS_CTL_POWER_CYCLE;
   else if (strncmp(argv[0], "reset", 5) == 0)
      ctl = IPMI_CHASSIS_CTL_HARD_RESET;
   else if (strncmp(argv[0], "diag", 4) == 0)
      ctl = IPMI_CHASSIS_CTL_PULSE_DIAG;
   else if ((strncmp(argv[0], "acpi", 4) == 0) || (strncmp(argv[0], "soft", 4) == 0))
      ctl = IPMI_CHASSIS_CTL_ACPI_SOFT;
   else {
      lprintf(LOG_ERR, "Invalid chassis power command: %s", argv[0]);
      return -1;
   }

   rc = ipmi_chassis_power_control(intf, ctl);
   return rc;
}

int
ipmi_chassis_main(struct ipmi_intf * intf, int argc, char ** argv)
{
   int rc = 0;

   if ((argc == 0) || (strncmp(argv[0], "help", 4) == 0)) {
   	lprintf(LOG_NOTICE, "Chassis Commands:  status, power, identify");

   }
   else if (strncmp(argv[0], "status", 6) == 0) {
      rc = ipmi_chassis_status(intf);
   }

   else if (strncmp(argv[0], "power", 5) == 0) {
      uint8_t ctl = 0;

      if ((argc < 2) || (strncmp(argv[1], "help", 4) == 0)) {
         lprintf(LOG_NOTICE, "chassis power Commands: status, on, off, cycle, reset, diag, soft");
         return 0;
      }
      if (strncmp(argv[1], "status", 6) == 0) {
         rc = ipmi_chassis_print_power_status(intf);
         return rc;
      }
      if ((strncmp(argv[1], "up", 2) == 0) || (strncmp(argv[1], "on", 2) == 0))
         ctl = IPMI_CHASSIS_CTL_POWER_UP;
      else if ((strncmp(argv[1], "down", 4) == 0) || (strncmp(argv[1], "off", 3) == 0))
         ctl = IPMI_CHASSIS_CTL_POWER_DOWN;
      else if (strncmp(argv[1], "cycle", 5) == 0)
         ctl = IPMI_CHASSIS_CTL_POWER_CYCLE;
      else if (strncmp(argv[1], "reset", 5) == 0)
         ctl = IPMI_CHASSIS_CTL_HARD_RESET;
      else if (strncmp(argv[1], "diag", 4) == 0)
         ctl = IPMI_CHASSIS_CTL_PULSE_DIAG;
      else if ((strncmp(argv[1], "acpi", 4) == 0) || (strncmp(argv[1], "soft", 4) == 0))
         ctl = IPMI_CHASSIS_CTL_ACPI_SOFT;
      else {
         lprintf(LOG_ERR, "Invalid chassis power command: %s", argv[1]);
         return -1;
      }

      rc = ipmi_chassis_power_control(intf, ctl);
   }
   else if (strncmp(argv[0], "identify", 8) == 0) {
      if (argc < 2) {
         rc = ipmi_chassis_identify(intf, NULL);
      }
      else if (strncmp(argv[1], "help", 4) == 0) {
         lprintf(LOG_NOTICE, "chassis identify <interval>");
         lprintf(LOG_NOTICE, "                 default is 15 seconds");
         lprintf(LOG_NOTICE, "                 0 to turn off");
         lprintf(LOG_NOTICE, "                 force to turn on indefinitely");
      } else {
         rc = ipmi_chassis_identify(intf, argv[1]);
      }
   }

   else {
      lprintf(LOG_ERR, "Invalid Chassis command: %s", argv[0]);
      return -1;
   }

   return rc;
}
